/* quaternion/gsl_quaternion_math.h
 * 
 * Copyright (C) 2006 Frederic-Emmanuel PICCA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __GSL_QUATERNION_MATH_H__
#define __GSL_QUATERNION_MATH_H__
#include <gsl/gsl_quaternion.h>
#include <gsl/gsl_vector.h>

#undef __BEGIN_DECLS
#undef __END_DECLS
#ifdef __cplusplus
#define __BEGIN_DECLS extern "C" {
#define __END_DECLS }
#else
#define __BEGIN_DECLS           /* empty */
#define __END_DECLS             /* empty */
#endif

__BEGIN_DECLS

/* Quaternion numbers */

gsl_quaternion gsl_quaternion_rect (double a, double b, double c, double d);  /* q = real + i imag + j jmag + k kmag */
gsl_quaternion gsl_quaternion_polar (double theta, gsl_vector * v); /* q = cos(theta/2) + sin(theta/2)*(i v[0] + j v[1] + k v[2]) */

#define GSL_QUATERNION_ONE (gsl_quaternion_rect(1.0,0.0,0.0,0.0))
#define GSL_QUATERNION_ZERO (gsl_quaternion_rect(0.0,0.0,0.0,0.0))
#define GSL_QUATERNION_NEGONE (gsl_quaternion_rect(-1.0,0.0,0.0,0.0))

#ifdef HAVE_INLINE
extern inline gsl_quaternion
gsl_quaternion_rect (double a, double b, double c, double d)
{                               /* return z = a + i b + j c + k d*/
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, a, b, c, d);
  return q;
}

extern inline gsl_quaternion
gsl_quaternion_polar(double theta, gsl_vector * v)
{
    if (v->size != 3)
      { 
        GSL_ERROR_VAL ("The v vector must be a 3D vector", GSL_EBADLEN, GSL_QUATERNION_ZERO);
      }
    else
      {
        double v0 = gsl_vector_get(v, 0);
        double v1 = gsl_vector_get(v, 1);
        double v2 = gsl_vector_get(v, 2);
        double n = sqrt(v0*v0 + v1*v1 + v2*v2);
        
        double theta_2 = theta / 2.;
        double c = cos(theta_2);
        double s = sin(theta_2) / n;
        
        gsl_quaternion q;
        GSL_SET_QUATERNION (&q, c, (s * v0), (s * v1), (s * v2));
        return q;
      }
}
#endif // HAVE_INLINE

/* Properties of quaternion numbers */

double gsl_quaternion_arg (gsl_quaternion z); /* return arg(z), -pi< arg(z) <=+pi */
double gsl_quaternion_abs (gsl_quaternion z);   /* return |z|   */
double gsl_quaternion_abs2 (gsl_quaternion z);  /* return |z|^2 */

/* quaternion arithmetic operators */

gsl_quaternion gsl_quaternion_add (gsl_quaternion a, gsl_quaternion b);  /* q=a+b */
gsl_quaternion gsl_quaternion_sub (gsl_quaternion a, gsl_quaternion b);  /* q=a-b */
gsl_quaternion gsl_quaternion_mul (gsl_quaternion a, gsl_quaternion b);  /* q=a*b */
                                                           
gsl_quaternion gsl_quaternion_add_real (gsl_quaternion a, double x);  /* q=a+x */
gsl_quaternion gsl_quaternion_sub_real (gsl_quaternion a, double x);  /* q=a-x */
gsl_quaternion gsl_quaternion_mul_real (gsl_quaternion a, double x);  /* q=a*x */
gsl_quaternion gsl_quaternion_div_real (gsl_quaternion a, double x);  /* q=a/x */

gsl_quaternion gsl_quaternion_add_imag (gsl_quaternion a, double y);  /* q=a+iy */
gsl_quaternion gsl_quaternion_sub_imag (gsl_quaternion a, double y);  /* q=a-iy */
gsl_quaternion gsl_quaternion_mul_imag (gsl_quaternion a, double y);  /* q=a*iy */

gsl_quaternion gsl_quaternion_add_jmag (gsl_quaternion a, double y);  /* q=a+jy */
gsl_quaternion gsl_quaternion_sub_jmag (gsl_quaternion a, double y);  /* q=a-jy */
gsl_quaternion gsl_quaternion_mul_jmag (gsl_quaternion a, double y);  /* q=a*jy */

gsl_quaternion gsl_quaternion_add_kmag (gsl_quaternion a, double y);  /* q=a+ky */
gsl_quaternion gsl_quaternion_sub_kmag (gsl_quaternion a, double y);  /* q=a-ky */
gsl_quaternion gsl_quaternion_mul_kmag (gsl_quaternion a, double y);  /* q=a*ky */

gsl_quaternion gsl_quaternion_conjugate (gsl_quaternion a);  /* q=conj(a) */
gsl_quaternion gsl_quaternion_neg (gsl_quaternion a);   /* q=-a */
gsl_quaternion gsl_quaternion_invert (gsl_quaternion a);    /* q=1/a */

__END_DECLS

#endif /* __GSL_COMPLEX_MATH_H__ */
