/* quaternion/math.c
 * 
 * Copyright (C) 2006 Frdric-Emmanuel PICCA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <config.h>
#include <math.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_quaternion.h>
#include <gsl/gsl_quaternion_math.h>

/**********************************************************************
 * Quaternion numbers
 **********************************************************************/

#ifndef HIDE_INLINE_STATIC
gsl_quaternion
gsl_quaternion_rect (double a, double b, double c, double d)
{                               /* return q = a + i b + j c + k d */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, a, b, c, d);
  return q;
}

gsl_quaternion
gsl_quaternion_polar (double theta, gsl_vector * v)
{
    if (v->size != 3)
      { 
        GSL_ERROR_VAL ("The v vector must be a 3D vector", GSL_EBADLEN, GSL_QUATERNION_ZERO);
      }
    else
      {
        double v0 = gsl_vector_get(v, 0);
        double v1 = gsl_vector_get(v, 1);
        double v2 = gsl_vector_get(v, 2);
        double n = sqrt(v0*v0 + v1*v1 + v2*v2);
        
        double theta_2 = theta / 2.;
        double c = cos (theta_2);
        double s = sin (theta_2) / n;
        
        gsl_quaternion q;
        GSL_SET_QUATERNION (&q, (c), (s * v0), (s * v1), (s * v2));
        return q;
      }
}
#endif

/**********************************************************************
 * Properties of quaternion numbers
 **********************************************************************/

double
gsl_quaternion_arg (gsl_quaternion q)
{                               /* return arg(q),  0 < arg(q) <= +pi */
  double x = GSL_QUATERNION_REAL (q);
  double y = gsl_quaternion_abs (q);

  if (y == 0.0)
    {
      return 0;
    }

  return acos (x/y);
}

double
gsl_quaternion_abs (gsl_quaternion q)
{                               /* return |q| */
  double a = GSL_QUATERNION_REAL (q);
  double b = GSL_QUATERNION_IMAG (q); 
  double c = GSL_QUATERNION_JMAG (q);
  double d = GSL_QUATERNION_KMAG (q);
  
  return sqrt (a * a + b * b + c * c + d * d);
}

double
gsl_quaternion_abs2 (gsl_quaternion q)
{                               /* return |q|^2 */
  double a = GSL_QUATERNION_REAL (q);
  double b = GSL_QUATERNION_IMAG (q);
  double c = GSL_QUATERNION_JMAG (q);
  double d = GSL_QUATERNION_KMAG (q);
  
  return (a * a + b * b + c * c + d * d);
}

/***********************************
 * Quaternion arithmetic operators *
 ***********************************/

gsl_quaternion
gsl_quaternion_add (gsl_quaternion a, gsl_quaternion b)
{                               /* q=a+b */
  double aa = GSL_QUATERNION_REAL (a), ab = GSL_QUATERNION_IMAG (a), ac = GSL_QUATERNION_JMAG(a), ad = GSL_QUATERNION_KMAG(a);
  double ba = GSL_QUATERNION_REAL (b), bb = GSL_QUATERNION_IMAG (b), bc = GSL_QUATERNION_JMAG(b), bd = GSL_QUATERNION_KMAG(b);

  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, aa + ba, ab + bb, ac + bc, ad + bd);
  return q;
}

gsl_quaternion
gsl_quaternion_add_real (gsl_quaternion a, double x)
{                               /* q=a+x */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a) + x, GSL_QUATERNION_IMAG (a), GSL_QUATERNION_JMAG (a), GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_add_imag (gsl_quaternion a, double x)
{                               /* q=a+ix */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a), GSL_QUATERNION_IMAG (a) + x, GSL_QUATERNION_JMAG (a), GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_add_jmag (gsl_quaternion a, double x)
{                               /* q=a+jx */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a), GSL_QUATERNION_IMAG (a), GSL_QUATERNION_JMAG (a) + x, GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_add_kmag (gsl_quaternion a, double x)
{                               /* q=a+kx */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a), GSL_QUATERNION_IMAG (a), GSL_QUATERNION_JMAG (a), GSL_QUATERNION_KMAG (a) + x);
  return q;
}


gsl_quaternion
gsl_quaternion_sub (gsl_quaternion a, gsl_quaternion b)
{                               /* q=a-b */
  double aa = GSL_QUATERNION_REAL (a), ab = GSL_QUATERNION_IMAG (a), ac = GSL_QUATERNION_JMAG(a), ad = GSL_QUATERNION_KMAG(a);
  double ba = GSL_QUATERNION_REAL (b), bb = GSL_QUATERNION_IMAG (b), bc = GSL_QUATERNION_JMAG(b), bd = GSL_QUATERNION_KMAG(b);

  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, aa - ba, ab - bb, ac - bc, ad - bd);
  return q;
}

gsl_quaternion
gsl_quaternion_sub_real (gsl_quaternion a, double x)
{                               /* z=a-x */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a) - x, GSL_QUATERNION_IMAG (a), GSL_QUATERNION_JMAG (a), GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_sub_imag (gsl_quaternion a, double x)
{                               /* z=a-ix */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a), GSL_QUATERNION_IMAG (a) - x, GSL_QUATERNION_JMAG (a), GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_sub_jmag (gsl_quaternion a, double x)
{                               /* z=a-jx */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a), GSL_QUATERNION_IMAG (a), GSL_QUATERNION_JMAG (a) - x, GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_sub_kmag (gsl_quaternion a, double x)
{                               /* z=a-kx */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, GSL_QUATERNION_REAL (a), GSL_QUATERNION_IMAG (a), GSL_QUATERNION_JMAG (a), GSL_QUATERNION_KMAG (a) - x);
  return q;
}

gsl_quaternion
gsl_quaternion_mul (gsl_quaternion a, gsl_quaternion b)
{                               /* q=a*b */
  double aa = GSL_QUATERNION_REAL (a), ab = GSL_QUATERNION_IMAG (a), ac = GSL_QUATERNION_JMAG(a), ad = GSL_QUATERNION_KMAG(a);
  double ba = GSL_QUATERNION_REAL (b), bb = GSL_QUATERNION_IMAG (b), bc = GSL_QUATERNION_JMAG(b), bd = GSL_QUATERNION_KMAG(b);

  double qa = aa * ba - ab * bb - ac * bc - ad * bd;
  double qb = ab * ba + aa * bb + ac * bd - ad * bc;
  double qc = ac * ba + aa * bc + ad * bb - ab * bd;
  double qd = ad * ba + aa * bd + ab * bc - ac * bb;
  
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, qa, qb, qc, qd); 
  return q;
}

gsl_quaternion
gsl_quaternion_mul_real (gsl_quaternion a, double x)
{                               /* q=a*x */
  double aa = GSL_QUATERNION_REAL (a), ab = GSL_QUATERNION_IMAG (a), ac = GSL_QUATERNION_JMAG(a), ad = GSL_QUATERNION_KMAG(a);
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, x * GSL_QUATERNION_REAL (a), x * GSL_QUATERNION_IMAG (a), x * GSL_QUATERNION_JMAG (a), x * GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_mul_imag (gsl_quaternion a, double x)
{                               /* q=a*ix */
  double aa = GSL_QUATERNION_REAL (a), ab = GSL_QUATERNION_IMAG (a), ac = GSL_QUATERNION_JMAG(a), ad = GSL_QUATERNION_KMAG(a);
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, -ab * x, aa * x, ad * x, -ac * x);
  return q;
}

gsl_quaternion
gsl_quaternion_mul_jmag (gsl_quaternion a, double x)
{                               /* q=a*jx */
  double aa = GSL_QUATERNION_REAL (a), ab = GSL_QUATERNION_IMAG (a), ac = GSL_QUATERNION_JMAG(a), ad = GSL_QUATERNION_KMAG(a);
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, -ac * x, -ad * x, aa * x, ab * x);
  return q;
}

gsl_quaternion
gsl_quaternion_mul_kmag (gsl_quaternion a, double x)
{                               /* q=a*kx */
  double aa = GSL_QUATERNION_REAL (a), ab = GSL_QUATERNION_IMAG (a), ac = GSL_QUATERNION_JMAG(a), ad = GSL_QUATERNION_KMAG(a);
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q, -ad * x, ac * x, -ab * x, aa * x);
  return q;
}

gsl_quaternion
gsl_quaternion_div_real (gsl_quaternion a, double x)
{                               /* q=a/x */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q,
                      GSL_QUATERNION_REAL (a) / x,
                      GSL_QUATERNION_IMAG (a) / x,
                      GSL_QUATERNION_JMAG (a) / x,
                      GSL_QUATERNION_KMAG (a) / x);
  return q;
}

gsl_quaternion
gsl_quaternion_conjugate (gsl_quaternion a)
{                               /* z=conj(a) */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q,
                      GSL_QUATERNION_REAL (a),
                      -GSL_QUATERNION_IMAG (a),
                      -GSL_QUATERNION_JMAG (a),
                      -GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_neg (gsl_quaternion a)
{                               /* z=-a */
  gsl_quaternion q;
  GSL_SET_QUATERNION (&q,
                      -GSL_QUATERNION_REAL (a),
                      -GSL_QUATERNION_IMAG (a),
                      -GSL_QUATERNION_JMAG (a),
                      -GSL_QUATERNION_KMAG (a));
  return q;
}

gsl_quaternion
gsl_quaternion_invert (gsl_quaternion a)
{                               /* q=(a*)/(a.a*) */
  double s = 1. / gsl_quaternion_abs2 (a);

  gsl_quaternion q;
  GSL_SET_QUATERNION (&q,
                      GSL_QUATERNION_REAL (a) * s,
                      -(GSL_QUATERNION_IMAG (a) * s),
                      -(GSL_QUATERNION_JMAG (a) * s),
                      -(GSL_QUATERNION_KMAG (a) * s));
  return q;
}
